@@ -31,11 +31,15 @@
  * Version 2.  See the file COPYING for more details.
  */
 
+#include "config.h"
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
 #include <ctype.h>
 
+#ifdef UNITTEST
+#undef ENABLE_NLS
+#endif
 #include "e2fsck.h"
 
 #include "crc32defs.h"
@@ -83,10 +87,10 @@ __u32 crc32_le(__u32 crc, unsigned char const *p, size_t len)
 	const __u32      *b =(__u32 *)p;
 	const __u32      *tab = crc32table_le;
 
-# ifdef __LITTLE_ENDIAN
-#  define DO_CRC(x) crc = tab[ (crc ^ (x)) & 255 ] ^ (crc>>8)
-# else
+# ifdef WORDS_BIGENDIAN
 #  define DO_CRC(x) crc = tab[ ((crc >> 24) ^ (x)) & 255] ^ (crc<<8)
+# else
+#  define DO_CRC(x) crc = tab[ (crc ^ (x)) & 255 ] ^ (crc>>8)
 # endif
 
 	crc = __cpu_to_le32(crc);
@@ -183,10 +187,10 @@ __u32 crc32_be(__u32 crc, unsigned char const *p, size_t len)
 	const __u32      *b =(const __u32 *)p;
 	const __u32      *tab = crc32table_be;
 
-# ifdef __LITTLE_ENDIAN
-#  define DO_CRC(x) crc = tab[ (crc ^ (x)) & 255 ] ^ (crc>>8)
-# else
+# ifdef WORDS_BIGENDIAN
 #  define DO_CRC(x) crc = tab[ ((crc >> 24) ^ (x)) & 255] ^ (crc<<8)
+# else
+#  define DO_CRC(x) crc = tab[ (crc ^ (x)) & 255 ] ^ (crc>>8)
 # endif
 
 	crc = __cpu_to_be32(crc);
